"""
Created 18th June 2019
"""
from urllib.parse import urlsplit, parse_qs


class LogLineProcessor:
    """
    Class for applying log processing on a line by line basis.
    Cleansing of PID data is applied here. Specific keys can
    be passed to further processor classes if additional
    processing is required.
    """

    def __init__(self):
        self.processors = {}
        self.processor_keys = []

        self.pid_references = self.make_sensitive_substitutions()
        self.setprocessor_keys()

    @staticmethod
    def make_sensitive_substitutions():
        """
        List of substitution references - where patient identifiable data
        should be stripped from Operation Logs. Substitution is case
        insensitive.
        """
        # Force all keys to lower - they are compared against key.lower()
        return set(
            x.lower()
            for x in [
                "nhsNumber",
                "successorNhsNumber",
                "motherNhsNumber",
                "babyNhsNumber",
                # The following lines are output in Demographics Trace Logging
                "queryLine",
                "givenName",
                "familyName",
                # The following lines are output in CPIS Logging.
                "authIdentityNo",
                "laCode",
                "laName",
                "cpoType",
            ]
        )

    def setprocessor_keys(self):
        """
        List of keys which are subject to additional processing
        """
        self._add_processor("url", self.url_key_handler)
        self._add_processor("requestUrl", self.url_key_handler)

    def _add_processor(self, key, handler):
        """
        Add a processor and corresponding handler instance
        to the log processor.
        """
        lower_key = key.lower()
        self.processor_keys.append(lower_key)
        self.processors[lower_key] = handler

    # pylint: disable=broad-except
    # We really do want to catch a general exception in process().
    # There is scope for failures in the log key handlers which
    # would cause critical application errors. Since all logging
    # may pass through this method, validating all input edge cases
    # is difficult. It's safer to catch the general exception and
    # carry on as normal.
    def process(self, key, log_line):
        """
        Determine if we should process this type of key.
        If so, pass processing to the appropriate handler.
        """
        audit_row_required = False
        if not key:
            return audit_row_required

        lower = key.lower()

        if lower in self.pid_references:
            audit_row_required = True

        elif lower in self.processor_keys:
            try:
                audit_row_required = self.processors[lower](log_line)
            except Exception:
                pass

        return audit_row_required

    def url_key_handler(self, url):
        """
        Handler for url key types. Remove PID which
        has been passed as a URL parameter
        """
        query = urlsplit(url.lower())[3]
        params = parse_qs(query)

        return len(set(params.keys()) & self.pid_references) > 0


def check_for_param_dictionary(param_dict):
    """
    If key is 'parameters' and value is a dict, check the keys within the dict
    for any which need to be in audit.
    """
    if isinstance(param_dict, dict):
        for param_key in param_dict:
            audit = LogLineProcessor().process(param_key, param_dict[param_key])
            if audit:
                return True

    return False



"""
Created 18th June 2019
"""
import re
import six

from spine_aws_common.log.loglineprocessor import (
    check_for_param_dictionary,
    LogLineProcessor,
)


NOT_PROVIDED = "NotProvided"

DEFAULT_KEYS = ["internalID", "sessionid"]


def _default_key(log_row_dict, key, value):
    """
    If not already set, puts key=value into the dictionary: primarily useful
    for preventing key substitution errors for common fields.
    """
    if key not in log_row_dict:
        log_row_dict[key] = value


def _substitute_if_empty(value):
    """
    Splunk SDK version being used doesn't handle empty text nodes efficiently
    when parsing the results of a splunk query. This method substitutes any
    empty strings with text NotProvided.
    """
    if isinstance(value, six.string_types) and not value.strip():
        return NOT_PROVIDED
    return value


def add_default_keys(log_row_dict):
    """
    Add any default keys
    """
    for key in DEFAULT_KEYS:
        _default_key(log_row_dict, key, "")


def create_log_preamble(log_level, process_name, log_reference):
    """Creates the string to form the initial part of any log message"""
    return f"Log_Level={log_level} Process={process_name} logReference={log_reference}"


def can_encode_string(value):
    """
    Python 2 and 3 compatible check to see if value is either a base or
    unicode string.
    """
    return isinstance(value, str)


def _decode_unicode_dictionary(substitution_dict):
    """
    Resolve any unicode issues with the dictionary
    """
    decode_dict = {}
    for key in substitution_dict:
        if can_encode_string(substitution_dict[key]):
            decode_dict[key] = substitution_dict[key].encode("utf-8")
        else:
            decode_dict[key] = substitution_dict[key]
    return decode_dict


def evaluate_log_keys(log_details, log_row_dict):
    """
    Evaulate each log key to ensure no empty logs slip through
    """
    for log_key in log_row_dict:
        audit = check_for_param_dictionary(log_row_dict[log_key])
        log_details.audit_log_required = audit or log_details.audit_log_required

        audit = LogLineProcessor().process(log_key, log_row_dict[log_key])
        log_details.audit_log_required = audit or log_details.audit_log_required

        log_row_dict[log_key] = _substitute_if_empty(log_row_dict[log_key])


def create_log_line(log_preamble, log_text, substitution_dict):
    """
    Write a log line, catching error scenarios (unicode and missing terms in
    dictionary) and ensuring everything fits on a single line.
    """
    try:
        log_line = log_preamble + " - " + log_text.format(**substitution_dict)
    except KeyError as err:
        log_line = log_preamble + " - " + log_text + " - "
        log_line += "No substitution due to KeyError, missing keys: "
        log_line += str(list(err.args))
        log_line += ", dictionary of "
        log_line += str(substitution_dict)
        print("Substitution failure - fail build: " + log_line)
    except UnicodeError:
        decode_dict = _decode_unicode_dictionary(substitution_dict)
        log_line = log_preamble + " - " + log_text.format(**decode_dict)

    # Replace newlines with spaces
    log_line = re.sub(r"\n|\r", " ", log_line)
    return log_line


def substitute_preamble_for_monitor(log_preamble):
    """
    Switch the Log Level for the monitor version of info/audit log
    """
    if "Log_Level=AUDIT" in log_preamble:
        return log_preamble.replace("Log_Level=AUDIT", "Log_Level=MONITOR")
    if "Log_Level=INFO" in log_preamble:
        return log_preamble.replace("Log_Level=INFO", "Log_Level=MONITOR")
    return log_preamble



"""
Module for MESH API functionality for step functions
"""
from http import HTTPStatus
import os

from spine_aws_common import LambdaApplication

from mesh_aws_client.mesh_common import MeshCommon, SingletonCheckFailure
from mesh_aws_client.mesh_mailbox import MeshMailbox


class MeshPollMailboxApplication(LambdaApplication):
    """
    MESH API Lambda for sending a message
    """

    def __init__(self, additional_log_config=None, load_ssm_params=False):
        """
        Init variables
        """
        super().__init__(additional_log_config, load_ssm_params)
        self.mailbox_name = None
        self.environment = os.environ.get("Environment", "default")
        self.get_messages_step_function_name = self.system_config.get(
            "GET_MESSAGES_STEP_FUNCTION_NAME", f"{self.environment}-get-messages"
        )

    def initialise(self):
        # initialise
        self.mailbox_name = self.event["mailbox"]

    def start(self):
        # in case of crash
        self.response = {"statusCode": HTTPStatus.INTERNAL_SERVER_ERROR.value}

        try:
            MeshCommon.singleton_check(
                self.mailbox_name,
                self.get_messages_step_function_name,
            )
        except SingletonCheckFailure as e:
            self.response = MeshCommon.return_failure(
                self.log_object,
                HTTPStatus.TOO_MANY_REQUESTS.value,
                "MESHPOLL0003",
                self.mailbox_name,
                message=e.msg,
            )
            return

        mailbox = MeshMailbox(self.log_object, self.mailbox_name, self.environment)
        list_response, message_list = mailbox.list_messages()
        list_response.raise_for_status()
        message_count = len(message_list)
        output_list = []
        if message_count == 0:
            # return 204 to keep state transitions to minimum if no messages
            self.response = {"statusCode": HTTPStatus.NO_CONTENT.value, "body": {}}
            return
        for message in message_list:
            output_list.append(
                {
                    "headers": {"Content-Type": "application/json"},
                    "body": {
                        "complete": False,
                        "internal_id": self.log_object.internal_id,
                        "message_id": message,
                        "dest_mailbox": self.mailbox_name,
                    },
                }
            )

        # to set response for the lambda
        self.response = {
            "statusCode": HTTPStatus.OK.value,
            "headers": {"Content-Type": "application/json"},
            "body": {
                "internal_id": self.log_object.internal_id,
                "message_count": message_count,
                "message_list": output_list,
            },
        }


# create instance of class in global space
# this ensures initial setup of logging/config is only done on cold start
app = MeshPollMailboxApplication()


def lambda_handler(event, context):
    """Standard lambda_handler"""
    return app.main(event, context)



"""
Base S3 Event Lambda application
"""
from aws_lambda_powertools.utilities.data_classes import S3Event
from spine_aws_common.batch_application import BatchApplication


class S3EventApplication(BatchApplication):
    """
    Base class for S3 Event Lambda applications
    """

    EVENT_TYPE = S3Event

    def _get_internal_id_from_record(self, record):
        """
        Always create new internalID for S3 Events
        """
        return self._create_new_internal_id()



"""
Base SQS Lambda application
"""
from aws_lambda_powertools.utilities.data_classes import SQSEvent
from spine_aws_common.batch_application import BatchApplication


class SQSApplication(BatchApplication):
    """
    Base class for SQS Lambda applications
    """

    EVENT_TYPE = SQSEvent

    def _get_internal_id_from_record(self, record):
        """
        Get (or create new) internalID from record
        """
        if "internal_id" in record.message_attributes:
            return record.message_attributes["internal_id"]["stringValue"]
        return self._create_new_internal_id()
